#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Disable warnings on strange function names
# pylint: disable=C0103


import collections

import nose.tools
from nose.tools import assert_equal

from paleomix.common.text import \
    TableError, \
    padded_table, \
    parse_padded_table, \
    parse_lines, \
    parse_lines_by_contig


###############################################################################
###############################################################################
# Tests for 'padded_table'

def _padded_table(*args, **kwargs):
    return list(padded_table(*args, **kwargs))


def test_padded_table__empty():
    assert_equal(_padded_table(()), [])


def test_padded_table__single_line():
    table = [(1, 20, 3000)]
    expected = ["1    20    3000"]
    assert_equal(expected, _padded_table(table))


def test_padded_table__two_lines():
    table = [(1, 20, 3000),
             (3000, 20, 1)]
    expected = ["1       20    3000",
                "3000    20    1"]
    assert_equal(expected, _padded_table(table))


def test_padded_table__three_lines():
    table = [(1, 20, 3000),
             (3000, 20, 1),
             (1, 2, 30)]
    expected = ["1       20    3000",
                "3000    20    1",
                "1       2     30"]
    assert_equal(expected, _padded_table(table))


def test_padded_table__with_text():
    row_1, line_1 = (1, 20, 3000), "1       20    3000"
    row_2, line_2 = (3000, 20, 1), "3000    20    1"
    comment = "# An insightful comment goes here"

    def _do_test_padded_table__padding__with_text(table, expected):
        assert_equal(expected, _padded_table(table))
    yield _do_test_padded_table__padding__with_text, \
        [comment, row_1, row_2], [comment, line_1, line_2]
    yield _do_test_padded_table__padding__with_text, \
        [row_1, comment, row_2], [line_1, comment, line_2]
    yield _do_test_padded_table__padding__with_text, \
        [row_1, row_2, comment], [line_1, line_2, comment]


###############################################################################
###############################################################################
# Tests for 'parse_padded_table'

def _parse_padded_table(*args, **kwargs):
    return list(parse_padded_table(*args, **kwargs))


def test_parse_padded_table__empty():
    assert_equal([], _parse_padded_table([]))


def test_parse_padded_table__header_only():
    assert_equal([], _parse_padded_table(["A  B  C  D"]))


def test_parse_padded_table__single_row():
    table = ["A    B    C    D",
             "4    3    2    1"]
    expected = [{"A": '4', "B": '3', "C": '2', "D": '1'}]
    assert_equal(expected, _parse_padded_table(table))


def test_parse_padded_table__two_rows():
    table = ["A     B     C     D",
             "4     3     2     1",
             "AB    CD    EF    GH"]
    expected = [{"A": '4', "B": '3', "C": '2', "D": '1'},
                {"A": "AB", "B": "CD", "C": "EF", "D": "GH"}]
    assert_equal(expected, _parse_padded_table(table))


# Any amount of whitespace is allowed
def test_parse_padded_table__single_row__with_whitespace():
    table = ["A   B    C       E F",
             "1        0  1    2   3"]
    expected = [{"A": '1', "B": '0', "C": '1', "E": '2', "F": '3'}]
    assert_equal(expected, _parse_padded_table(table))


# Other whitespace should be ignored
def test_parse_padded_table__single_row__with_tabs():
    table = ["A\t\t\t\tB",
             "1\t\t\t\t0"]
    expected = [{"A": '1', "B": '0'}]
    assert_equal(expected, _parse_padded_table(table))


def test_padded_table__comments_and_empty_lines():
    def _do_test_padded_table__comments(lines):
        expected = [{"A": '3000', "B": '20', "C": '1'}]
        assert_equal(expected, _parse_padded_table(lines))
    line_1 = "A         B      C"
    line_2 = "3000      20      1"

    yield _do_test_padded_table__comments, ["# comment", line_1, line_2]
    yield _do_test_padded_table__comments, [line_1, " # comment", line_2]
    yield _do_test_padded_table__comments, [line_1, line_2, "  # comment"]

    yield _do_test_padded_table__comments, ["", line_1, line_2]
    yield _do_test_padded_table__comments, [line_1, "  ", line_2]
    yield _do_test_padded_table__comments, [line_1, line_2, "   "]


def test_padded_table__newlines():
    expected = [{"A": '3000', "B": '20', "C": '1'}]

    def _do_test_padded_table__padding__comments(postfix):
        line_1 = "A         B       C" + postfix
        line_2 = "3000      20      1" + postfix
        assert_equal(expected, _parse_padded_table([line_1, line_2]))

    yield _do_test_padded_table__padding__comments, "\r"
    yield _do_test_padded_table__padding__comments, "\n"
    yield _do_test_padded_table__padding__comments, "\r\n"


def test_padded_table__padding__comments__whitespace():
    expected = [{"A": '3000', "B": '20', "C": '1'}]
    lines = ["A         B       C",
             "3000      20      1",
             "  # useless comment"]
    assert_equal(expected, _parse_padded_table(lines))


@nose.tools.raises(TableError)
def test_parse_padded_table__malformed_table_0():
    table = ["A    B    C    D",
             "4    3    2"]
    _parse_padded_table(table)


@nose.tools.raises(TableError)
def test_parse_padded_table__malformed_table_1():
    table = ["A    B    C    D",
             "4    3    2    1    0"]
    _parse_padded_table(table)


###############################################################################
###############################################################################
# Tests for 'parse_linse'
def _this(*args):
    return args


def _parse_lines(*args, **kwargs):
    return list(parse_lines(*args, **kwargs))


def test_parse_lines__empty_file():
    def _assert_false():
        assert False  # pragma: no coverage
    assert_equal(_parse_lines([], _assert_false), [])


def test_parse_lines__single():
    assert_equal(_parse_lines(["abc line1 \n"], _this), [("abc line1", 9)])


def test_parse_lines__comments_and_empty_lines():
    def _do_test_parse_lines__comments(lines):
        expected = [("abc line1", 9), ("def line2", 9)]
        assert_equal(_parse_lines(lines, _this), expected)
    yield _do_test_parse_lines__comments, \
        ["# comment\n", "abc line1 \n", "def line2 \n"]
    yield _do_test_parse_lines__comments, \
        ["abc line1 \n", " # comment\n", "def line2 \n"]
    yield _do_test_parse_lines__comments, \
        ["abc line1 \n", "def line2 \n", "   # comment\n"]

    yield _do_test_parse_lines__comments, \
        ["\n", "abc line1 \n", "def line2 \n"]
    yield _do_test_parse_lines__comments, \
        ["abc line1 \n", " \n", "def line2 \n"]
    yield _do_test_parse_lines__comments, \
        ["abc line1 \n", "def line2 \n", "   \n"]


def test_parse_lines__padding__newlines():
    expected = [("abc line1", 9), ("def line2", 9)]

    def _do_test_padded_table__padding__comments(postfix):
        line_1 = "abc line1 " + postfix
        line_2 = "def line2 " + postfix
        assert_equal(expected, _parse_lines([line_1, line_2], _this))

    yield _do_test_padded_table__padding__comments, "\r"
    yield _do_test_padded_table__padding__comments, "\n"
    yield _do_test_padded_table__padding__comments, "\r\n"


@nose.tools.raises(TypeError)
def test_parse_lines__uncallable():
    _parse_lines([], 1)


###############################################################################
###############################################################################
# Tests for 'parse_lines_by_contig'
_RecordMock = collections.namedtuple("_RecordMock", "contig value")


def test_parse_lines_by_contig__single_contig():
    lines = ["abc line1 \n", "abc line2 \n"]

    def _parse(line, length):
        assert_equal(len(line), length)
        return _RecordMock(*line.split())

    expected = {"abc": [_RecordMock("abc", "line1"),
                        _RecordMock("abc", "line2")]}
    assert_equal(parse_lines_by_contig(lines, _parse), expected)


def test_parse_lines__two_contigs():
    lines = ["abc line1 \n", "def line2 \n"]

    def _parse(line, length):
        assert_equal(len(line), length)
        return _RecordMock(*line.split())

    expected = {"abc": [_RecordMock("abc", "line1")],
                "def": [_RecordMock("def", "line2")]}
    assert_equal(parse_lines_by_contig(lines, _parse), expected)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# pylint: disable=missing-docstring
#
import os
import signal

import nose
from nose.tools import assert_equal
from paleomix.common.testing import with_temp_folder

from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import ParallelCmds, SequentialCmds
from paleomix.atomiccmd.pprint import pformat, _pformat_list


###############################################################################
###############################################################################
# pformat

def test_pformat__simple():
    cmd = AtomicCmd(("touch", "something"))
    assert_equal(pformat(cmd),
                 ("Command = touch something\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'")
                 % (id(cmd), id(cmd)))


@with_temp_folder
def test_pformat__simple__running(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    assert_equal(pformat(cmd),
                 ("Command = sleep 10\n"
                  "Status  = Running ...\n"
                  "STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
                  "STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
                  "CWD     = '{cwd}'").format(id=id(cmd),
                                              cwd=os.getcwd(),
                                              temp_dir=temp_folder))
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_pformat__simple__running__set_cwd(temp_folder):
    cmd = AtomicCmd(("sleep", "10"), set_cwd=True)
    cmd.run(temp_folder)
    assert_equal(pformat(cmd),
                 ("Command = sleep 10\n"
                  "Status  = Running ...\n"
                  "STDOUT* = 'pipe_sleep_{id}.stdout'\n"
                  "STDERR* = 'pipe_sleep_{id}.stderr'\n"
                  "CWD     = '{temp_dir}'").format(id=id(cmd),
                                                   temp_dir=temp_folder))
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_pformat__simple__done(temp_folder):
    cmd = AtomicCmd("true")
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert_equal(pformat(cmd),
                 ("Command = true\n"
                  "Status  = Exited with return-code 0\n"
                  "STDOUT* = '{temp_dir}/pipe_true_{id}.stdout'\n"
                  "STDERR* = '{temp_dir}/pipe_true_{id}.stderr'\n"
                  "CWD     = '{cwd}'").format(id=id(cmd),
                                              cwd=os.getcwd(),
                                              temp_dir=temp_folder))


@with_temp_folder
def test_pformat__simple__done__before_join(temp_folder):
    cmd = AtomicCmd("true")
    cmd.run(temp_folder)
    # pylint: disable=protected-access
    cmd._proc.wait()
    assert_equal(pformat(cmd),
                 ("Command = true\n"
                  "Status  = Exited with return-code 0\n"
                  "STDOUT* = '{temp_dir}/pipe_true_{id}.stdout'\n"
                  "STDERR* = '{temp_dir}/pipe_true_{id}.stderr'\n"
                  "CWD     = '{cwd}'").format(id=id(cmd),
                                              cwd=os.getcwd(),
                                              temp_dir=temp_folder))
    assert_equal(cmd.join(), [0])


@with_temp_folder
def test_pformat__simple__done__set_cwd(temp_folder):
    cmd = AtomicCmd("true", set_cwd=True)
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert_equal(pformat(cmd),
                 ("Command = true\n"
                  "Status  = Exited with return-code 0\n"
                  "STDOUT* = 'pipe_true_{id}.stdout'\n"
                  "STDERR* = 'pipe_true_{id}.stderr'\n"
                  "CWD     = '{temp_dir}'").format(id=id(cmd),
                                                   temp_dir=temp_folder))


@with_temp_folder
def test_pformat__simple__terminated_by_pipeline(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    cmd.terminate()
    assert_equal(cmd.join(), ["SIGTERM"])
    assert_equal(pformat(cmd),
                 ("Command = sleep 10\n"
                  "Status  = Automatically terminated by PALEOMIX\n"
                  "STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
                  "STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
                  "CWD     = '{cwd}'").format(id=id(cmd),
                                              temp_dir=temp_folder,
                                              cwd=os.getcwd()))


@with_temp_folder
def test_pformat__simple__killed_by_signal(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    # pylint: disable=protected-access
    os.killpg(cmd._proc.pid, signal.SIGTERM)
    assert_equal(cmd.join(), ["SIGTERM"])
    assert_equal(pformat(cmd),
                 ("Command = sleep 10\n"
                  "Status  = Terminated with signal SIGTERM\n"
                  "STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
                  "STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
                  "CWD     = '{cwd}'").format(id=id(cmd),
                                              temp_dir=temp_folder,
                                              cwd=os.getcwd()))


@with_temp_folder
def test_pformat__simple__temp_root_in_arguments(temp_folder):
    cmd = AtomicCmd(("echo", "${TEMP_DIR}"))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert_equal(pformat(cmd),
                 ("Command = echo '{temp_dir}'\n"
                  "Status  = Exited with return-code 0\n"
                  "STDOUT* = '{temp_dir}/pipe_echo_{id}.stdout'\n"
                  "STDERR* = '{temp_dir}/pipe_echo_{id}.stderr'\n"
                  "CWD     = '{cwd}'")
                 .format(id=id(cmd),
                         temp_dir=temp_folder,
                         cwd=os.getcwd()))


###############################################################################
###############################################################################
# INFILE

def test_pformat__atomiccmd__simple_with_infile():
    cmd = AtomicCmd(("cat", "%(IN_SOMETHING)s"), IN_SOMETHING="/etc/fstab")
    assert_equal(pformat(cmd),
                 ("Command = cat /etc/fstab\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_cat_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_cat_%i.stderr'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_infile__set_cwd():
    cmd = AtomicCmd(("cat", "%(IN_SOMETHING)s"),
                    IN_SOMETHING="/etc/fstab",
                    set_cwd=True)
    assert_equal(pformat(cmd),
                 ("Command = cat /etc/fstab\n"
                  "STDOUT* = 'pipe_cat_%i.stdout'\n"
                  "STDERR* = 'pipe_cat_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_infile():
    cmd = AtomicCmd(("cat", "%(TEMP_IN_FILE)s"),
                    TEMP_IN_FILE="infile.txt")

    assert_equal(pformat(cmd),
                 ("Command = cat '${TEMP_DIR}/infile.txt'\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_cat_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_cat_%i.stderr'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_infile__set_cwd():
    cmd = AtomicCmd(("zcat", "%(TEMP_IN_FILE)s"),
                    TEMP_IN_FILE="infile.gz",
                    set_cwd=True)

    assert_equal(pformat(cmd),
                 ("Command = zcat infile.gz\n"
                  "STDOUT* = 'pipe_zcat_%i.stdout'\n"
                  "STDERR* = 'pipe_zcat_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'")
                 % (id(cmd), id(cmd)))


###############################################################################
###############################################################################
# OUTFILE

def test_pformat__atomiccmd__simple_with_outfile():
    cmd = AtomicCmd(("touch", "%(OUT_RC)s"), OUT_RC="/etc/bashrc")
    assert_equal(pformat(cmd),
                 ("Command = touch '${TEMP_DIR}/bashrc'\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_outfile__set_cwd():
    cmd = AtomicCmd(("touch", "%(OUT_RC)s"),
                    OUT_RC="/etc/bashrc",
                    set_cwd=True)

    assert_equal(pformat(cmd),
                 ("Command = touch bashrc\n"
                  "STDOUT* = 'pipe_touch_%i.stdout'\n"
                  "STDERR* = 'pipe_touch_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_outfile():
    cmd = AtomicCmd(("touch", "%(TEMP_OUT_RC)s"),
                    TEMP_OUT_RC="bashrc")

    assert_equal(pformat(cmd),
                 ("Command = touch '${TEMP_DIR}/bashrc'\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_outfile__set_cwd():
    cmd = AtomicCmd(("touch", "%(TEMP_OUT_RC)s"),
                    TEMP_OUT_RC="bashrc",
                    set_cwd=True)

    assert_equal(pformat(cmd),
                 ("Command = touch bashrc\n"
                  "STDOUT* = 'pipe_touch_%i.stdout'\n"
                  "STDERR* = 'pipe_touch_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'")
                 % (id(cmd), id(cmd)))


###############################################################################
###############################################################################
# STDIN

def test_pformat__atomiccmd__simple_with_stdin():
    cmd = AtomicCmd("gzip", IN_STDIN="/etc/fstab")
    assert_equal(pformat(cmd),
                 ("Command = gzip\n"
                  "STDIN   = '/etc/fstab'\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_stdin__set_cwd():
    cmd = AtomicCmd("gzip", IN_STDIN="/etc/fstab", set_cwd=True)
    assert_equal(pformat(cmd),
                 ("Command = gzip\n"
                  "STDIN   = '/etc/fstab'\n"
                  "STDOUT* = 'pipe_gzip_%i.stdout'\n"
                  "STDERR* = 'pipe_gzip_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_stdin():
    cmd = AtomicCmd("gzip", TEMP_IN_STDIN="stabstabstab")
    assert_equal(pformat(cmd),
                 ("Command = gzip\n"
                  "STDIN*  = '${TEMP_DIR}/stabstabstab'\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_stdin__set_cwd():
    cmd = AtomicCmd("gzip", TEMP_IN_STDIN="stabstabstab", set_cwd=True)
    assert_equal(pformat(cmd),
                 ("Command = gzip\n"
                  "STDIN*  = 'stabstabstab'\n"
                  "STDOUT* = 'pipe_gzip_%i.stdout'\n"
                  "STDERR* = 'pipe_gzip_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'")
                 % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_stdin__cmd():
    cmd_1 = AtomicCmd("gzip", OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
    assert_equal(pformat(cmd_2),
                 ("Command = gzip\n"
                  "STDIN   = <PIPE>\n"
                  "STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'")
                 % (id(cmd_2), id(cmd_2)))


###############################################################################
###############################################################################
# STDOUT

def test_pformat__atomiccmd__simple_with_stdout():
    cmd = AtomicCmd(("echo", "Water. Water."), OUT_STDOUT="/dev/ls")
    assert_equal(pformat(cmd),
                 ("Command = echo 'Water. Water.'\n"
                  "STDOUT  = '${TEMP_DIR}/ls'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'")
                 % (id(cmd),))


def test_pformat__atomiccmd__simple_with_stdout__set_cwd():
    cmd = AtomicCmd(("echo", "*pant*. *pant*."),
                    OUT_STDOUT="/dev/barf", set_cwd=True)
    assert_equal(pformat(cmd),
                 ("Command = echo '*pant*. *pant*.'\n"
                  "STDOUT  = 'barf'\n"
                  "STDERR* = 'pipe_echo_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'") % (id(cmd),))


def test_pformat__atomiccmd__simple_with_temp_stdout():
    cmd = AtomicCmd(("echo", "Oil. Oil."), TEMP_OUT_STDOUT="dm")
    assert_equal(pformat(cmd),
                 ("Command = echo 'Oil. Oil.'\n"
                  "STDOUT* = '${TEMP_DIR}/dm'\n"
                  "STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'")
                 % (id(cmd),))


def test_pformat__atomiccmd__simple_with_temp_stdout__set_cwd():
    cmd = AtomicCmd(("echo", "Room service. Room service."),
                    TEMP_OUT_STDOUT="pv",
                    set_cwd=True)
    assert_equal(pformat(cmd),
                 ("Command = echo 'Room service. Room service.'\n"
                  "STDOUT* = 'pv'\n"
                  "STDERR* = 'pipe_echo_%i.stderr'\n"
                  "CWD     = '${TEMP_DIR}'") % (id(cmd),))


def test_pformat__atomiccmd__simple_with_stdout_pipe():
    cmd = AtomicCmd(("echo", "!"), OUT_STDOUT=AtomicCmd.PIPE)
    assert_equal(pformat(cmd),
                 ("Command = echo '!'\n"
                  "STDOUT  = <PIPE>\n"
                  "STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'")
                 % (id(cmd),))


###############################################################################
###############################################################################
# ParallelCmds

def test_pformat__sets__simple():
    template = ("{description}:\n"
                "  Process 1:\n"
                "    Command = echo foo\n"
                "    STDOUT  = Piped to process 2\n"
                "    STDERR* = '${{TEMP_DIR}}/pipe_echo_{cmd_1_id}.stderr'\n"
                "\n"
                "  Process 2:\n"
                "    Command = gzip\n"
                "    STDIN   = Piped from process 1\n"
                "    STDOUT* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stdout'\n"
                "    STDERR* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stderr'")

    def _test_pformat__sets__simple(cls, description):
        cmd_1 = AtomicCmd(("echo", "foo"), OUT_STDOUT=AtomicCmd.PIPE)
        cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
        cmd = cls((cmd_1, cmd_2))
        assert_equal(pformat(cmd),
                     template.format(description=description,
                                     cmd_1_id=id(cmd_1),
                                     cmd_2_id=id(cmd_2)))

    yield _test_pformat__sets__simple, ParallelCmds, "Parallel processes"
    yield _test_pformat__sets__simple, SequentialCmds, "Sequential processes"


def test_pformat__sets__nested():
    cmd_1 = AtomicCmd(("echo", "foo"), OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
    cmd_3 = AtomicCmd("sha1sum")
    set_1 = ParallelCmds((cmd_1, cmd_2))
    set_2 = SequentialCmds((set_1, cmd_3))
    assert_equal(pformat(set_2),
                 ("Sequential processes:\n"
                  "  Parallel processes:\n"
                  "    Process 1:\n"
                  "      Command = echo foo\n"
                  "      STDOUT  = Piped to process 2\n"
                  "      STDERR* = '${{TEMP_DIR}}/pipe_echo_{cmd_1}.stderr'\n"
                  "\n"
                  "    Process 2:\n"
                  "      Command = gzip\n"
                  "      STDIN   = Piped from process 1\n"
                  "      STDOUT* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2}.stdout'\n"
                  "      STDERR* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2}.stderr'\n"
                  "\n"
                  "  Process 3:\n"
                  "    Command = sha1sum\n"
                  "    STDOUT* = '${{TEMP_DIR}}/pipe_sha1sum_{cmd_3}.stdout'\n"
                  "    STDERR* = '${{TEMP_DIR}}/pipe_sha1sum_{cmd_3}.stderr'")
                 .format(cmd_1=id(cmd_1),
                         cmd_2=id(cmd_2),
                         cmd_3=id(cmd_3)))


###############################################################################
###############################################################################
# Bad input

def test_pformat__bad_input():
    @nose.tools.raises(TypeError)
    def _do_test_pformat__bad_input(value):
        pformat(value)
    yield _do_test_pformat__bad_input, 1
    yield _do_test_pformat__bad_input, {}
    yield _do_test_pformat__bad_input, ""


###############################################################################
###############################################################################
# _pformat_list

def test_pformat_list__empty():
    assert_equal(_pformat_list([]), "")


def test_pformat_list__single():
    assert_equal(_pformat_list([3]), "3")


def test_pformat_list__multiple():
    assert_equal(_pformat_list([3, 2, 1]), "3 2 1")


def test_pformat_list__wrapped():
    assert_equal(_pformat_list([3, 2, 1], width=1), "3 \\\n    2 \\\n    1")
    assert_equal(_pformat_list([3, 2, 1], width=2), "3 \\\n    2 \\\n    1")
    assert_equal(_pformat_list([3, 2, 1], width=3), "3 \\\n    2 \\\n    1")
    assert_equal(_pformat_list([3, 2, 1], width=4), "3 2 \\\n    1")
    assert_equal(_pformat_list([3, 2, 1], width=5), "3 2 \\\n    1")
    assert_equal(_pformat_list([3, 2, 1], width=6), "3 2 1")
    assert_equal(_pformat_list([3, 2, 1], width=7), "3 2 1")


def test_pformat_list__escaped():
    assert_equal(_pformat_list(["a", "b c"], width=100), "a 'b c'")
    assert_equal(_pformat_list(["a", "$c"], width=100), "a '$c'")
    assert_equal(_pformat_list(["!a", "c"], width=100), "'!a' c")
    assert_equal(_pformat_list(["a", "'c"], width=100), """a ''"'"'c'""")



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import json
import math
import numbers
import os
import re
import sys

from paleomix.node import Node, NodeError
from paleomix.common.utilities import set_in, get_in
from paleomix.common.fileutils import move_file, reroot_path
from paleomix.tools.bam_stats.coverage import \
    read_table as read_coverage_table
from paleomix.common.bedtools import BEDRecord

import paleomix.common.text as text


_PE_READS = frozenset(("Paired", "Singleton",
                       "Collapsed", "CollapsedTruncated"))
_SE_READS = frozenset(("Single",))


class SummaryTableNode(Node):
    def __init__(self, config, makefile, target, cov_for_lanes, cov_for_libs, dependencies = ()):
        self._target        = target.name
        self._output_file   = os.path.join(config.destination, self._target + ".summary")
        self._prefixes      = makefile["Prefixes"]
        self._makefile      = makefile["Statistics"]

        self._in_raw_bams = cov_for_lanes
        self._in_lib_bams = cov_for_libs
        input_files = set()
        input_files.update(sum(map(list, self._in_raw_bams.values()), []))
        input_files.update(sum(map(list, self._in_lib_bams.values()), []))

        self._in_raw_read = collections.defaultdict(list)
        for prefix in target.prefixes:
            for sample in prefix.samples:
                for library in sample.libraries:
                    for lane in library.lanes:
                        filename = None
                        filetype = None

                        if lane.reads.stats:
                            filetype = 'Raw'
                            filename = lane.reads.stats
                        elif lane.reads.validation:
                            filename = lane.reads.validation
                            fileset = set(lane.reads.files)

                            if fileset & _PE_READS and fileset & _SE_READS:
                                filetype = '*'
                            elif fileset & _PE_READS:
                                filetype = 'PE'
                            elif fileset & _SE_READS:
                                filetype = 'SE'
                            else:
                                assert False, lane.reads.files
                                continue

                            input_files.add(filename)
                        else:
                            assert False

                        self._in_raw_read[(sample.name, library.name, lane.name)] = (filetype, filename)

        Node.__init__(self,
                      description  = "<Summary: %s>" % self._output_file,
                      input_files  = filter(None, input_files),
                      output_files = [self._output_file],
                      dependencies = dependencies)


    def _run(self, config, temp):
        rois    = self._stat_areas_of_interest(self._prefixes)
        genomes = self._stat_prefixes(self._prefixes)
        with open(reroot_path(temp, self._output_file), "w") as table:
            table.write("# Command:\n")
            table.write("#     %s\n" % (" ".join(sys.argv)),)
            table.write("#\n")
            table.write("# Directory:\n")
            table.write("#     %s\n" % (os.getcwd()),)
            table.write("#\n")
            table.write("# Makefile:\n")
            table.write("#     Filename: %s\n" % (self._makefile["Filename"],))
            table.write("#     SHA1Sum:  %s\n" % (self._makefile["Hash"],))
            table.write("#     MTime:    %s\n" % (self._makefile["MTime"],))
            table.write("#\n")
            self._write_genomes(table, genomes)
            table.write("#\n")
            self._write_areas_of_interest(table, rois)
            table.write("#\n#\n")

            for roi in rois.itervalues():
                genomes[roi["Label"]] = {"Size" : roi["Size"]}
            self._write_tables(table, genomes)


    def _teardown(self, _config, temp):
        move_file(reroot_path(temp, self._output_file), self._output_file)


    def _write_genomes(self, table, genomes):
        table.write("# Genomes:\n")
        rows = [["Name", "Label", "Contigs", "Size", "Prefix"]]
        for (_, prefix) in sorted(self._prefixes.items()):
            stats = genomes[prefix["Name"]]
            rows.append((prefix["Name"], prefix.get("Label", "-"), stats["NContigs"], stats["Size"], prefix["Path"]))

        for line in text.padded_table(rows):
            table.write("#     %s\n" % (line,))


    def _write_areas_of_interest(self, table, rois):
        table.write("# Regions Of Interest:\n")
        rows = [["Genome", "ROI", "Size", "NFeatures", "NIntervals", "Path"]]
        for (_, roi) in sorted(rois.items()):
            rows.append([roi[key] for key in ("Genome", "Name", "Size", "NFeatures", "NIntervals", "Path")])

        for line in text.padded_table(rows):
            table.write("#     %s\n" % (line,))


    def _write_tables(self, out, genomes):
        for row in self._build_tables(genomes):
            out.write("%s\n" % '\t'.join(map(str, row)))

    def _build_tables(self, genomes):
        yield ["Target", "Sample", "Library", "Measure", "Value", "# Description"]

        for (target, samples) in sorted(self._read_tables(self._prefixes, genomes).iteritems()):
            for (sample, libraries) in sorted(samples.iteritems()):
                for (library, prefixes) in sorted(libraries.iteritems()):
                    ordered = [("reads", prefixes.pop("reads"))] if "reads" in prefixes else []
                    ordered.extend(sorted(prefixes.items()))

                    for (prefix, table) in ordered:
                        table.pop("hits_unique_nts(%s)" % prefix, None)

                        for (key, (value, comment)) in sorted(table.iteritems(), key = _measure_ordering):
                            if isinstance(value, numbers.Number) and math.isnan(value):
                                value = "NA"
                            yield (target, sample, library, key, value, comment)
                        yield ""
                    yield ""

    def _read_tables(self, prefixes, genomes):
        table = {}
        self._read_reads_settings(table)
        self._read_raw_bam_stats(table)
        self._read_lib_bam_stats(table)

        for (target, samples) in table.items():
            merged_samples = {}
            for (sample, libraries) in samples.items():
                merged_libraries = {}
                for (library, subtables) in libraries.items():
                    for (tblname, subtable) in subtables.items():
                        merged_libraries[tblname] = self._merge_tables((merged_libraries.get(tblname, {}), subtable))
                        merged_samples[tblname] = self._merge_tables((merged_samples.get(tblname, {}), subtable))
                    libraries[library] = self._annotate_subtables(subtables, genomes)
                set_in(table, (target, sample, "*"), self._annotate_subtables(merged_libraries, genomes))
            set_in(table, (target, "*", "*"), self._annotate_subtables(merged_samples, genomes))

        return table


    @classmethod
    def _annotate_subtables(cls, subtables, genomes):
        if "mitochondrial" in subtables and "nuclear" in subtables:
            subtables["endogenous"] = cls._create_endogenous_subtable(subtables, genomes)

        for (tblname, subtable) in subtables.iteritems():
            if tblname == "reads":
                fractions = [("seq_trash_se",      "seq_reads_se",       "seq_trash_se_frac",   "# Fraction of SE reads trashed"),
                             ("seq_trash_pe_1",    "seq_reads_pairs",    "seq_trash_pe_1_frac", "# Fraction of PE mate 1 reads trashed"),
                             ("seq_trash_pe_2",    "seq_reads_pairs",    "seq_trash_pe_2_frac", "# Fraction of PE mate 2 reads trashed"),
                             ("seq_collapsed",     "seq_reads_pairs",    "seq_collapsed_frac",  "# Fraction of PE pairs collapsed into one read"),
                             ("seq_retained_nts",  "seq_retained_reads", "seq_retained_length", "# Average number of NTs in retained reads")]

                for (numerator, denominator, measure, comment) in fractions:
                    if (numerator in subtable) and (denominator in subtable):
                        value = subtable[numerator][0] / float(subtable[denominator][0] or "NaN")
                        subtable[measure] = (value, comment)
            else:
                total_hits  = subtable["hits_raw(%s)" % tblname][0]
                total_nts   = subtable["hits_unique_nts(%s)" % tblname][0]
                total_uniq  = subtable["hits_unique(%s)" % tblname][0]
                total_reads = subtables.get("reads",{}).get("seq_retained_reads", (0,))[0]
                genome_size = genomes[tblname]["Size"]

                subtable["hits_raw_frac(%s)" % tblname] = (total_hits / float(total_reads or "NaN"), "# Total number of hits vs. total number of reads retained")
                subtable["hits_unique_frac(%s)" % tblname] = (total_uniq / float(total_reads or "NaN"), "# Total number of unique hits vs. total number of reads retained")
                subtable["hits_clonality(%s)" % tblname] = (1 - total_uniq / float(total_hits or "NaN"), "# Fraction of hits that were PCR duplicates")
                subtable["hits_length(%s)" % tblname] = (total_nts / float(total_uniq or "NaN"), "# Average number of aligned bases per unique hit")
                subtable["hits_coverage(%s)" % tblname] = (total_nts / float(genome_size or "NaN"), "# Estimated coverage from unique hits")

        return subtables


    @classmethod
    def _create_endogenous_subtable(self, subtables, genomes):
        nucl = subtables["nuclear"]
        mito = subtables["mitochondrial"]

        total_hits            = mito["hits_raw(mitochondrial)"][0]         + nucl["hits_raw(nuclear)"][0]
        total_hits_unique     = mito["hits_unique(mitochondrial)"][0]      + nucl["hits_unique(nuclear)"][0]
        total_hits_unique_nts = mito["hits_unique_nts(mitochondrial)"][0]  + nucl["hits_unique_nts(nuclear)"][0]

        ratio_hits, ratio_genome, ratio_genome_inv = "NA", "NA", "NA"
        if mito["hits_unique(mitochondrial)"][0]:
            ratio_nts    = float(nucl["hits_unique_nts(nuclear)"][0]) / mito["hits_unique_nts(mitochondrial)"][0]
            ratio_hits   = float(nucl["hits_unique(nuclear)"][0]) / mito["hits_unique(mitochondrial)"][0]
            ratio_genome = ratio_nts / ((float(genomes["nuclear"]["Size"]) * 2) / float(genomes["mitochondrial"]["Size"]))
            ratio_genome_inv = ratio_genome ** -1

        return {
            "hits_raw(endogenous)"         : (total_hits,                       "# Total number of hits against the nuclear and mitochondrial genome"),
            "hits_unique(endogenous)"      : (total_hits_unique,                "# Total number of unique reads (PCR duplicates removed)"),
            "hits_unique_nts(endogenous)"  : (total_hits_unique_nts,            None),
            "ratio_reads(nuc,mito)"        : (ratio_hits,                       "# Ratio of unique hits: Hits(nuc) / H(mito)"),
            "ratio_genome(nuc,mito)"       : (ratio_genome,                     "# Ratio of NTs of unique hits corrected by genome sizes: (NTs(nuc) / NTs(mito)) / ((2 * Size(nuc)) / Size(mito))"),
            "ratio_genome(mito,nuc)"       : (ratio_genome_inv,                 "# Ratio of NTs of unique hits corrected by genome sizes: (NTs(mito) / NTs(nuc)) / (Size(mito) / (2 * Size(nuc)))")
            }


    def _read_reads_settings(self, table):
        for ((sample, library, barcode), (filetype, filename)) in self._in_raw_read.iteritems():
            key = (self._target, sample, library, "reads", barcode)
            set_in(table, key, self._stat_read_settings(filetype, filename))

        for (_, samples) in table.iteritems():
            for (sample, libraries) in samples.iteritems():
                for (library, prefixes) in libraries.iteritems():
                    prefixes["reads"] = self._merge_tables(prefixes["reads"].values())

        return table


    def _read_raw_bam_stats(self, table):
        for ((genome, target, sample, library), filenames) in self._in_raw_bams.iteritems():
            key = (target, sample, library)
            hits, _ = self._read_coverage_tables(key, filenames)

            value = (hits, "# Total number of hits (prior to PCR duplicate filtering)")
            set_in(table, (target, sample, library, genome, "hits_raw(%s)" % genome), value)


    def _read_lib_bam_stats(self, table):
        for ((genome, target, sample, library), filenames) in self._in_lib_bams.iteritems():
            key = (target, sample, library)
            hits, nts = self._read_coverage_tables(key, filenames)

            value = (hits, "# Total number of hits (excluding any PCR duplicates)")
            set_in(table, (target, sample, library, genome, "hits_unique(%s)" % genome), value)
            set_in(table, (target, sample, library, genome, "hits_unique_nts(%s)" % genome), (nts, None))


    @classmethod
    def _read_coverage_tables(cls, key, filenames):
        hits = nts = 0
        for filename in filenames:
            subtable = {}
            read_coverage_table(subtable, filename)
            contigtables = get_in(subtable, key)

            if contigtables is None:
                raise NodeError("Error reading table %r; row not found:"
                                "\n   %s   ...\n\nIf files have been renamed "
                                "during the run, then please remove this file "
                                "in that it may be re-generated.\nHowever, "
                                "note that read-group tags in the BAM files "
                                "may not be correct!"
                                % (filename, "   ".join(key)))

            for contigtable in contigtables.itervalues():
                hits += contigtable["Hits"]
                nts += contigtable["M"]
        return hits, nts


    @classmethod
    def _merge_tables(cls, tables):
        merged = {}
        for table in tables:
            for (measure, (value, comment)) in table.iteritems():
                if not isinstance(value, numbers.Number):
                    other, _ = merged.get(measure, (value, None))
                    merged[measure] = (value if (value == other) else "*", comment)
                else:
                    other, _ = merged.get(measure, (0, None))
                    merged[measure] = (value + other, comment)
        return merged


    @classmethod
    def _stat_read_settings(cls, filetype, filename):
        assert filetype in ('Raw', 'SE', 'PE', '*', 'BAM'), filetype

        if filetype != 'Raw':
            stats = {
                "lib_type": (filetype, "# SE, PE, or * (for both)"),
                "seq_retained_nts": [float("nan"), "# Total number of NTs in retained reads"],
                "seq_retained_reads": [float("nan"), "# Total number of retained reads"],
            }

            if filetype in ('SE', '*'):
                stats["seq_reads_se"] = [float("nan"),  "# Total number of single-ended reads"]
                stats["seq_trash_se"] = [float("nan"),  "# Total number of trashed reads"]

            if filetype in ('PE', '*'):
                stats["seq_reads_pairs"] = [float("nan"), "# Total number of reads"]
                stats["seq_trash_pe_1"] = [float("nan"), "# Total number of reads"]
                stats["seq_trash_pe_2"] = [float("nan"), "# Total number of reads"]
                stats["seq_collapsed"] = [float("nan"), "# Total number of pairs collapsed into one read"]

            if filename is not None:
                with open(filename) as handle:
                    for key, value in json.load(handle).items():
                        if key in stats:
                            stats[key][0] = int(value)

            return stats

        with open(filename) as settings_file:
            settings = settings_file.read()
            def _re_search(regexp, default = None):
                match = re.search(regexp, settings)
                if not match:
                    if default is not None:
                        return default
                    raise KeyError("Could not find match with RegExp %s in file '%s'" \
                                   % (repr(regexp), filename))

                return int(match.groups()[0])

            if "Paired end mode" in settings or "paired-end reads" in settings:
                return {
                    "lib_type"            : ("PE", "# SE, PE, or * (for both)"),
                    "seq_reads_pairs"     : (_re_search("number of read pairs: ([0-9]+)"),    "# Total number of pairs"),
                    "seq_trash_pe_1"      : (_re_search("discarded mate 1 reads: ([0-9]+)"),  "# Total number of reads"),
                    "seq_trash_pe_2"      : (_re_search("discarded mate 2 reads: ([0-9]+)"),  "# Total number of reads"),
                    "seq_retained_nts"    : (_re_search("retained nucleotides: ([0-9]+)"),    "# Total number of NTs in retained reads"),
                    "seq_retained_reads"  : (_re_search("retained reads: ([0-9]+)"),          "# Total number of retained reads"),
                    "seq_collapsed"       : (_re_search("of (?:full-length )?collapsed pairs: ([0-9]+)", 0) + \
                                             _re_search("of truncated collapsed pairs: ([0-9]+)", 0),
                                            "# Total number of pairs collapsed into one read"),
                    }
            elif "Single end mode" in settings or "single-end reads" in settings:
                return {
                    "lib_type"            : ("SE", "# SE, PE, or * (for both)"),
                    "seq_reads_se"        : (_re_search("number of (?:reads|read pairs): ([0-9]+)"),  "# Total number of single-ended reads"),
                    "seq_trash_se"        : (_re_search("discarded mate 1 reads: ([0-9]+)"),          "# Total number of trashed reads"),
                    "seq_retained_nts"    : (_re_search("retained nucleotides: ([0-9]+)"),            "# Total number of NTs in retained reads"),
                    "seq_retained_reads"  : (_re_search("retained reads: ([0-9]+)"),                  "# Total number of retained reads"),
                    }
            else:
                assert False, filename


    @classmethod
    def _stat_areas_of_interest(cls, prefixes):
        """Returns (size, number of named intervals, total number of intervals)
        for a set of areas of interest."""
        areas_of_interest = {}
        for (prefix_name, prefix) in prefixes.iteritems():
            prefix_label = prefix.get("Label", prefix_name)
            for (roi_name, roi_filename) in prefix.get("RegionsOfInterest", {}).iteritems():
                count, names, size = 0, set(), 0
                with open(roi_filename) as handle:
                    for line in handle:
                        bed = BEDRecord(line)
                        names.add(bed.name if len(bed) >= 4 else (bed.contig + "*"))
                        size += (bed.end - bed.start)
                        count += 1
                areas_of_interest[(prefix_name, roi_name)] = {"Size"       : size,
                                                              "NFeatures"  : len(names),
                                                              "NIntervals" : count,
                                                              "Genome"     : prefix["Name"],
                                                              "Name"       : roi_name,
                                                              "Label"      : "%s:%s" % (prefix_label, roi_name),
                                                              "Path"       : roi_filename}
        return areas_of_interest


    @classmethod
    def _stat_prefixes(cls, prefixes):
        """Returns (size, number of contigs) for a set of BWA prefix."""
        genomes = {}
        for prefix in prefixes:
            with open(prefixes[prefix]["Reference"] + ".fai") as table:
                lengths = [int(line.split()[1]) for line in table]

            labels = [prefix]
            if "Label" in prefixes[prefix]:
                labels.append(prefixes[prefix].get("Label"))

            for label in labels:
                if label not in genomes:
                    genomes[label] = {"Size" : 0, "NContigs" : 0}

                statistics = genomes[label]
                statistics["Size"] += sum(lengths)
                statistics["NContigs"] += len(lengths)

        if "mitochondrial" in genomes and "nuclear" in genomes:
            nucl = genomes["nuclear"]
            mito = genomes["mitochondrial"]

            genomes["endogenous"] = {"Size" : nucl["Size"] + mito["Size"],
                                     "NContigs" : nucl["NContigs"] + mito["NContigs"]}

        return genomes



def _measure_ordering(pair):
    measure = pair[0]
    key = measure.split("(")[0]
    return (__ORDERING[key], measure)


__ORDERING = {
    "lib_type"             :  00,
    "seq_reads_se"         :  10,
    "seq_trash_se"         :  20,
    "seq_trash_se_frac"    :  30,
    "seq_reads_pairs"      :  40,
    "seq_trash_pe_1"       :  50,
    "seq_trash_pe_1_frac"  :  60,
    "seq_trash_pe_2"       :  70,
    "seq_trash_pe_2_frac"  :  80,
    "seq_collapsed"        :  90,
    "seq_collapsed_frac"   : 100,
    "seq_retained_reads"   : 110,
    "seq_retained_nts"     : 120,
    "seq_retained_length"  : 130,


    "hits_raw"             : 140,
    "hits_raw_frac"        : 150,
    "hits_clonality"       : 160,
    "hits_unique"          : 170,
    "hits_unique_frac"     : 180,
    "hits_coverage"        : 190,
    "hits_length"          : 200,
    "ratio_reads"          : 210,
    "ratio_genome"         : 220,
    }



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import itertools
import nose.tools
from nose.tools import assert_equal

from paleomix.common.sequences import \
     complement, \
     reverse_complement, \
     encode_genotype, \
     count_nts, \
     count_gc_diploid, \
     split


###############################################################################
###############################################################################
# Tests for 'complement'

_REF_SRC = "ACGTMRWSYKVHDBNX"
_REF_DST = "TGCAKYWSRMBDHVNX"
_REF_PAIRS = zip(_REF_SRC, _REF_DST)


def test_complement__single_nt():
    def test_function(source, destination):
        assert_equal(complement(source), destination)

    for (src, dest) in _REF_PAIRS:
        yield test_function, src, dest
        yield test_function, src.lower(), dest.lower()


def test_complement__multiple_nts_upper():
    assert_equal(complement(_REF_SRC), _REF_DST)


def test_complement__multiple_nts_lower():
    assert_equal(complement(_REF_SRC.lower()), _REF_DST.lower())


def test_complement__multiple_nts_mixed_case():
    assert_equal(complement("aGtCn"), "tCaGn")


###############################################################################
###############################################################################
# Tests for 'complement'

def test_reverse_complement():
    assert_equal(reverse_complement(_REF_SRC), _REF_DST[::-1])


###############################################################################
###############################################################################
# Tests for 'encode_genotype'

_IUB_SRC = ("A", "C", "G", "T",
            "AC", "AG", "AT", "CG", "CT", "GT",
            "ACG", "ACT", "AGT", "CGT", "ACGT")
_IUB_DST = "ACGTMRWSYKVHDB"
_IUB_PAIRS = zip(_IUB_SRC, _IUB_DST)


def test_genotype__permutations():
    def test_function(src, dst):
        assert_equal(encode_genotype(src), dst)

    for (src, dst) in _IUB_PAIRS:
        for seq in itertools.permutations(src):
            yield test_function, "".join(seq), dst


@nose.tools.raises(ValueError)
def test_genotype__bad_input__lowercase():
    encode_genotype("a")


@nose.tools.raises(ValueError)
def test_genotype__bad_input__mixedcase():
    encode_genotype("At")


@nose.tools.raises(ValueError)
def test_genotype__bad_input__unknown_nucleotide():
    encode_genotype("Z")


@nose.tools.raises(ValueError)
def test_genotype__bad_input__non_nucleotide():
    encode_genotype("+")


def test_comma_or_not():
    def test_function(sequence):
        assert_equal(encode_genotype(sequence), "Y")

    for sequence in ("CT", "C,T", ",C,T", "C,T,", ",C,T,"):
        yield test_function, sequence


###############################################################################
###############################################################################
# Tests for 'count_nts'

def test_count_nts__empty_str():
    assert_equal(count_nts(""), {})


def test_count_nts__simple():
    assert_equal(count_nts("TTATGTGTCT"), {"A": 1, "C": 1, "G": 2, "T": 6})


def test_count_nts__simple_mixed_case():
    assert_equal(count_nts("atATNcaGCC"), {"A": 3, "C": 3, "G": 1, "T": 2, "N": 1})


def test_count_nts__complex():
    assert_equal(count_nts("NNBKVGRVMM"), {"N": 2, "B": 1, "K": 1, "V": 2, "G": 1,
                                           "R": 1, "M": 2})


def test_count_nts__complex_mixed_case():
    assert_equal(count_nts(""), {})


@nose.tools.raises(ValueError)
def test_count_nts__bad_nt():
    count_nts("TTAPGTGTCT")  # 'P' not UIPAC


@nose.tools.raises(ValueError)
def test_count_nts__bad_nt_lowercase():
    count_nts("atATNcoGCC")  # 'o' not UIPAC


###############################################################################
###############################################################################
# Tests for 'count_gc_diploid'

def test_count_gc_diploid__empty_string():
    assert_equal(count_gc_diploid(""), (0, 0))


def test_count_gc_diploid__simple():
    assert_equal(count_gc_diploid("ACGGGTATCA"), (10, 20))


def test_count_gc_diploid__simple_mixed_case():
    assert_equal(count_gc_diploid("AcGGgtAtcA"), (10, 20))


def test_count_gc_diploid__complex():
    assert_equal(count_gc_diploid("RTGTKMGTCA"), (9, 20))


def test_count_gc_diploid__complex_mixed_case():
    assert_equal(count_gc_diploid("rtGTKMgTcA"), (9, 20))


def test_count_gc_dipoid__ambigious():
    assert_equal(count_gc_diploid("ACGGNTATCA"), (8, 18))


def test_count_gc_diploid__ambigous_mixed_case():
    assert_equal(count_gc_diploid("AcGGntAtcA"), (8, 18))


@nose.tools.raises(ValueError)
def test_count_gc_diploid__triploid():
    count_gc_diploid("B")


###############################################################################
###############################################################################
# Tests for 'split'

def test_split__empty_sequence():
    assert_equal(split(""), {"1": "", "2": "", "3": ""})


@nose.tools.raises(TypeError)
def test_split__no_split_by():
    split("", split_by="")


def test_split__single_group():
    assert_equal(split("ACGCAT", "111"), {'1': 'ACGCAT'})


def test_split__two_groups():
    assert_equal(split("ACGCAT", "112"), {"1": "ACCA", "2": "GT"})


def test_split__three_groups():
    expected = {"1": "AC", "2": "CA", "3": "GT"}
    assert_equal(split("ACGCAT", "123"), expected)
    assert_equal(split("ACGCAT"), expected)


def test_split__empty_group():
    expected = {"1": "A", "2": "C", "3": ""}
    assert_equal(split("AC"), expected)


def test_split__partial_group():
    expected = {"1": "AA", "2": "CA", "3": "G"}
    assert_equal(split("ACGAA"), expected)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import errno
import logging
import os

import paleomix.common.versions as versions

from paleomix.common.fileutils import \
    reroot_path, \
    missing_executables
from paleomix.common.utilities import \
    safe_coerce_to_frozenset


# Max number of error messages of each type
_MAX_ERROR_MESSAGES = 10


class FileStatusCache(object):
    """Cache used to avoid repeatedly checking the state (existance / mtime) of
    files required / generated by nodes. A new cache is generated for every
    operation (e.g. refreshing all states / manually setting the state of a
    node) to avoid relying on the filesystem staying consistant for long
    periods of time.
    """

    def __init__(self):
        self._stat_cache = {}

    def files_exist(self, fpaths):
        """Returns true if all paths listed in fpaths exist."""
        return all((self._get_state(fpath) is not None) for fpath in fpaths)

    def missing_files(self, fpaths):
        """Returns a list of paths in fpaths that do not exist."""
        return [fpath for fpath in fpaths if (self._get_state(fpath) is None)]

    def are_files_outdated(self, input_files, output_files):
        """Returns true if any 'input' files have a time-stamp that post-date
        any time-stamp for the 'output' files, indicating that one or more of
        the 'input' files have changed since the creation of the 'output'.

        The function also returns true if any files are missing, as this would
        indicate either the 'output' files or both the 'input' and 'output'
        files would need to be rebuilt.
        """
        input_timestamps = []
        if not self._get_states(input_files, input_timestamps):
            return True

        output_timestamps = []
        if not self._get_states(output_files, output_timestamps):
            return True

        return max(input_timestamps) > min(output_timestamps)

    def _get_states(self, filenames, dst):
        """Collects the mtimes for a set of filenames, returning true if all
        could be collected, and aborting early and returning false otherwise.
        """
        for filename in filenames:
            timestamp = self._get_state(filename)
            if timestamp is None:
                return False

            dst.append(timestamp)

        return True

    def _get_state(self, fpath):
        """Returns the mtime of a path, or None if the path does not exist."""
        if fpath not in self._stat_cache:
            try:
                mtime = os.path.getmtime(fpath)
            except OSError, error:
                if error.errno != errno.ENOENT:
                    raise
                mtime = None
            self._stat_cache[fpath] = mtime
        return self._stat_cache[fpath]


class NodeGraphError(RuntimeError):
    pass


class NodeGraph:
    NUMBER_OF_STATES = 6
    DONE, RUNNING, RUNABLE, QUEUED, OUTDATED, ERROR \
        = range(NUMBER_OF_STATES)

    def __init__(self, nodes, cache_factory=FileStatusCache):
        self._cache_factory = cache_factory
        self._state_observers = []
        self._states = {}

        nodes = safe_coerce_to_frozenset(nodes)

        self._logger = logging.getLogger(__name__)
        self._reverse_dependencies = collections.defaultdict(set)
        self._collect_reverse_dependencies(nodes, self._reverse_dependencies, set())
        self._intersections = {}
        self._top_nodes = [node for (node, rev_deps) in self._reverse_dependencies.iteritems() if not rev_deps]

        self._logger.info("  - Checking file dependencies ...")
        self._check_file_dependencies(self._reverse_dependencies)
        self._logger.info("  - Checking for required executables ...")
        self._check_required_executables(self._reverse_dependencies)
        self._logger.info("  - Checking version requirements ...")
        self._check_version_requirements(self._reverse_dependencies)
        self._logger.info("  - Determining states ...")
        self.refresh_states()
        self._logger.info("  - Ready ...\n")

    def get_node_state(self, node):
        return self._states[node]

    def set_node_state(self, node, state):
        if state not in (NodeGraph.RUNNING, NodeGraph.ERROR, NodeGraph.DONE):
            raise ValueError("Invalid state: %r" % (state,))
        old_state = self._states[node]
        if state == old_state:
            return

        self._states[node] = state
        self._notify_state_observers(node, old_state, state, True)

        intersections = self._calculate_intersections(node)

        # Not all nodes may need to be updated, but we still need to
        # traverse the "graph" (using the intersection counts) in order
        # to ensure that all nodes that need to be updated are updated.
        requires_update = dict.fromkeys(intersections, False)
        for dependency in self._reverse_dependencies[node]:
            requires_update[dependency] = True

        cache = self._cache_factory()
        while any(requires_update.itervalues()):
            for (node, count) in intersections.items():
                if not count:
                    has_changed = False
                    if requires_update[node]:
                        old_state = self._states.pop(node)
                        new_state = self._update_node_state(node, cache)
                        if new_state != old_state:
                            self._notify_state_observers(node, old_state,
                                                         new_state, False)
                            has_changed = True

                    for dependency in self._reverse_dependencies[node]:
                        intersections[dependency] -= 1
                        requires_update[dependency] |= has_changed

                    intersections.pop(node)
                    requires_update.pop(node)

    def __iter__(self):
        """Returns a graph of nodes."""
        return iter(self._top_nodes)

    def iterflat(self):
        return iter(self._reverse_dependencies)

    def refresh_states(self):
        states = {}
        cache = self._cache_factory()
        for (node, state) in self._states.iteritems():
            if state in (self.ERROR, self.RUNNING):
                states[node] = state
        self._states = states
        for node in self._reverse_dependencies:
            self._update_node_state(node, cache)
        self._refresh_state_observers()

    def add_state_observer(self, observer):
        """Add an observer of changes to the node-graph. The observer
        is expected to have the following functions:

        refresh(nodegraph):
          Called when an observer has been added, or when 'refresh_states'
          has been called on the nodegraph. The observer should rebuild any
          internal state at this point.

        state_changed(node, old_state, new_state, is_primary):
          Called when the state of an node has changed. 'is_primary' is
          True only if the node for which 'set_node_state' was called,
          and false for nodes the state of which changed as a consequence
          of the change to the node marked 'is_primary'. This includes
          ERROR propegating, and more."""
        self._state_observers.append(observer)
        observer.refresh(self)

    def _notify_state_observers(self, node, old_state, new_state, is_primary):
        for observer in self._state_observers:
            observer.state_changed(node, old_state, new_state, is_primary)

    def _refresh_state_observers(self):
        for observer in self._state_observers:
            observer.refresh(self)

    def _calculate_intersections(self, for_node):
        def count_nodes(node, counts):
            for node in self._reverse_dependencies[node]:
                if node in counts:
                    counts[node] += 1
                else:
                    counts[node] = 1
                    count_nodes(node, counts)
            return counts

        if for_node not in self._intersections:
            counts = count_nodes(for_node, {})
            for dependency in self._reverse_dependencies[for_node]:
                counts[dependency] -= 1
            self._intersections[for_node] = counts

        return dict(self._intersections[for_node])

    def _update_node_state(self, node, cache):
        if node in self._states:
            return self._states[node]

        # Update sub-nodes before checking for fixed states
        dependency_states = set((NodeGraph.DONE,))
        for dependency in node.dependencies:
            dependency_states.add(self._update_node_state(dependency, cache))

        state = max(dependency_states)
        if state == NodeGraph.DONE:
            if not self.is_done(node, cache):
                state = NodeGraph.RUNABLE
            elif not cache.files_exist(node.input_files):
                # Somehow the input files have gone missing, despite the
                # dependant nodes being done; this implies this node is
                # outdated, since the input-files should be re-generated, but
                # obviously it is not possible to run it at this point.
                missing = cache.missing_files(node.input_files)
                self._logger.error("ERROR: Input file(s) missing for node; "
                                   "may have been moved while the pipeline "
                                   "was running. Cannot proceed:\n"
                                   "    Node = %s\n    Files = %s\n"
                                   % (node, "\n            ".join(missing)))
                state = NodeGraph.ERROR
            elif self.is_outdated(node, cache):
                state = NodeGraph.RUNABLE
        elif state in (NodeGraph.RUNNING, NodeGraph.RUNABLE, NodeGraph.QUEUED):
            if self.is_done(node, cache):
                state = NodeGraph.OUTDATED
            else:
                state = NodeGraph.QUEUED
        self._states[node] = state

        return state

    @classmethod
    def is_done(cls, node, cache):
        """Returns true if the node itself is done; this only implies that the
        output files generated by this node exists. The files themselves may
        be outdated.
        """
        return cache.files_exist(node.output_files)

    @classmethod
    def is_outdated(cls, node, cache):
        """Returns true if the not is not done or if one or more of the input
        files appear to have been changed since the creation of the output
        files (based on the timestamps). A node that lacks either input or
        output files is never considered outdated.
        """
        if not (node.input_files and node.output_files):
            return False

        return cache.are_files_outdated(node.input_files, node.output_files)

    @classmethod
    def _check_required_executables(cls, nodes):
        exec_filenames = set()
        for node in nodes:
            exec_filenames.update(node.executables)

            # Requirements may include executables not invoked directly
            for requirement in node.requirements:
                if isinstance(requirement, versions.RequirementObj):
                    executable = requirement.executable
                    if executable is not None:
                        exec_filenames.add(executable)

        missing_exec = missing_executables(exec_filenames)
        if missing_exec:
            raise NodeGraphError("Required executables are missing:\n\t%s"
                                 % ("\n\t".join(sorted(missing_exec))))

    def _check_version_requirements(self, nodes):
        exec_requirements = set()
        for node in nodes:
            exec_requirements.update(node.requirements)

        def _key_func(reqobj):
            # Sort priority in decreasing order, name in increasing order
            return (-reqobj.priority, reqobj.name)
        exec_requirements = list(sorted(exec_requirements, key=_key_func))

        try:
            for requirement in exec_requirements:
                self._logger.info("    - Checking version of %r ..."
                                  % (requirement.name,))

                requirement()
        except versions.VersionRequirementError, error:
            raise NodeGraphError(error)
        except OSError, error:
            raise NodeGraphError("Could not check version for %s:\n\t%s"
                                 % (requirement.name, error))

    @classmethod
    def _check_file_dependencies(cls, nodes):
        files = ("input_files", "output_files")
        files = dict((key, collections.defaultdict(set)) for key in files)
        # Auxiliary files are treated as input files
        files["auxiliary_files"] = files["input_files"]

        for node in nodes:
            for (attr, nodes_by_file) in files.iteritems():
                for filename in getattr(node, attr):
                    nodes_by_file[filename].add(node)

        max_messages = range(_MAX_ERROR_MESSAGES)
        error_messages = []
        error_messages.extend(zip(max_messages, cls._check_output_files(files["output_files"])))
        error_messages.extend(zip(max_messages, cls._check_input_dependencies(files["input_files"],
                                                                              files["output_files"], nodes)))

        if error_messages:
            messages = []
            for (_, error) in error_messages:
                for line in error.split("\n"):
                    messages.append("\t" + line)

            raise NodeGraphError("Errors detected during graph construction (max %i shown):\n%s" \
                                % (_MAX_ERROR_MESSAGES * 2, "\n".join(messages)),)


    @classmethod
    def _check_output_files(cls, output_files):
        """Checks dict of output files to nodes for cases where
        multiple nodes create the same output file.

        The directory component of paths are realized in order to
        detect cases where nodes create the same file, but via
        different paths (e.g. due to relative/absolute paths, or
        due to use of symbolic links). Since output files are
        replaced, not modified in place, it is not nessesary to
        compare files themselves."""
        dirpath_cache, real_output_files = {}, {}
        for (filename, nodes) in output_files.iteritems():
            dirpath = os.path.dirname(filename)
            if dirpath not in dirpath_cache:
                dirpath_cache[dirpath] = os.path.realpath(dirpath)

            real_output_file = reroot_path(dirpath_cache[dirpath], filename)
            real_output_files.setdefault(real_output_file, []).extend(nodes)

        for (filename, nodes) in real_output_files.iteritems():
            if (len(nodes) > 1):
                nodes = _summarize_nodes(nodes)
                yield "Multiple nodes create the same (clobber) output-file:" \
                      "\n\tFilename: %s\n\tNodes: %s" \
                      % (filename, "\n\t       ".join(nodes))


    @classmethod
    def _check_input_dependencies(cls, input_files, output_files, nodes):
        dependencies = cls._collect_dependencies(nodes, {})

        for (filename, nodes) in sorted(input_files.items(), key = lambda v: v[0]):
            if (filename in output_files):
                producers = output_files[filename]
                bad_nodes = set()
                for consumer in nodes:
                    if not (producers & dependencies[consumer]):
                        bad_nodes.add(consumer)

                if bad_nodes:
                    producer  = iter(producers).next()
                    bad_nodes = _summarize_nodes(bad_nodes)
                    yield "Node depends on dynamically created file, but not on the node creating it:" + \
                                "\n\tFilename: %s\n\tCreated by: %s\n\tDependent node(s): %s" \
                                % (filename, producer, "\n\t                   ".join(bad_nodes))
            elif not os.path.exists(filename):
                nodes = _summarize_nodes(nodes)
                yield "Required file does not exist, and is not created by a node:" + \
                            "\n\tFilename: %s\n\tDependent node(s): %s" \
                            % (filename,    "\n\t                   ".join(nodes))

    @classmethod
    def _collect_dependencies(cls, nodes, dependencies):
        for node in nodes:
            if node not in dependencies:
                subnodes = node.dependencies
                if not subnodes:
                    dependencies[node] = frozenset()
                    continue

                cls._collect_dependencies(subnodes, dependencies)

                collected = set(subnodes)
                for subnode in subnodes:
                    collected.update(dependencies[subnode])
                dependencies[node] = frozenset(collected)

        return dependencies


    @classmethod
    def _collect_reverse_dependencies(cls, lst, rev_dependencies, processed):
        for node in lst:
            if node not in processed:
                processed.add(node)

                # Initialize default-dict
                rev_dependencies[node] # pylint: disable=W0104

                subnodes = node.dependencies
                for dependency in subnodes:
                    rev_dependencies[dependency].add(node)
                cls._collect_reverse_dependencies(subnodes, rev_dependencies, processed)



def _summarize_nodes(nodes):
    nodes = list(sorted(set(map(str, nodes))))
    if len(nodes) > 4:
        nodes = nodes[:5] + ["and %i more nodes ..." % len(nodes)]
    return nodes



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import time
import logging

import paleomix
import paleomix.logger
import paleomix.resources
import paleomix.yaml

from paleomix.common.console import \
    print_err, \
    print_info

from paleomix.pipeline import \
    Pypeline
from paleomix.nodes.picard import \
    BuildSequenceDictNode
from paleomix.nodes.samtools import \
    FastaIndexNode
from paleomix.nodes.bwa import \
    BWAIndexNode
from paleomix.nodes.bowtie2 import \
    Bowtie2IndexNode
from paleomix.nodes.validation import \
    ValidateFASTAFilesNode

from paleomix.tools.bam_pipeline.makefile import \
    MakefileError, \
    read_makefiles

from paleomix.tools.bam_pipeline.parts import \
    Reads

import paleomix.tools.bam_pipeline.parts as parts
import paleomix.tools.bam_pipeline.config as bam_config
import paleomix.tools.bam_pipeline.mkfile as bam_mkfile


def build_pipeline_trimming(config, makefile):
    """Builds only the nodes required to produce trimmed reads.
    This reduces the required complexity of the makefile to a minimum."""

    nodes = []
    for (_, samples) in makefile["Targets"].iteritems():
        print_info(".", end='')

        for libraries in samples.itervalues():
            for barcodes in libraries.itervalues():
                for record in barcodes.itervalues():
                    if record["Type"] in ("Raw", "Trimmed"):
                        offset = record["Options"]["QualityOffset"]
                        reads = Reads(config, record, offset)

                        nodes.extend(reads.nodes)

    return nodes


def build_pipeline_full(config, makefile, return_nodes=True):
    result = []
    features = makefile["Options"]["Features"]
    for (target_name, sample_records) in makefile["Targets"].iteritems():
        print_info(".", end='')

        prefixes = []
        for (_, prefix) in makefile["Prefixes"].iteritems():
            samples = []
            for (sample_name, library_records) in sample_records.iteritems():
                libraries = []
                for (library_name, barcode_records) in library_records.iteritems():
                    lanes = []
                    for (barcode, record) in barcode_records.iteritems():
                        lane = parts.Lane(config, prefix, record, barcode)

                        # ExcludeReads settings may exlude entire lanes
                        if lane.bams:
                            lanes.append(lane)

                    if lanes:
                        libraries.append(parts.Library(config=config,
                                                       target=target_name,
                                                       prefix=prefix,
                                                       lanes=lanes,
                                                       name=library_name))

                if libraries:
                    samples.append(parts.Sample(config=config,
                                                prefix=prefix,
                                                libraries=libraries,
                                                name=sample_name))

            if samples:
                prefixes.append(parts.Prefix(config=config,
                                             prefix=prefix,
                                             samples=samples,
                                             features=features,
                                             target=target_name))

        if prefixes:
            target = parts.Target(config, prefixes, target_name)

            # Construct coverage, depth-histogram, and summary nodes, etc.
            parts.add_statistics_nodes(config, makefile, target)

            if return_nodes:
                # Extra tasks (e.g. coverage, depth-histograms, etc.)
                result.extend(target.nodes)
                # Output BAM files (raw, realigned)
                result.extend(target.bams.itervalues())
            else:
                result.append(target)

    return result


def index_references(config, makefiles):
    references = {}
    references_bwa = {}
    references_bowtie2 = {}
    for makefile in makefiles:
        for subdd in makefile["Prefixes"].itervalues():
            reference = subdd["Reference"]
            if reference not in references:
                # Validation of the FASTA file; not blocking for the other
                # steps, as it is only expected to fail very rarely, but will
                # block subsequent analyses depending on the FASTA.
                valid_node = ValidateFASTAFilesNode(input_files=reference,
                                                    output_file=reference +
                                                    ".validated")
                # Indexing of FASTA file using 'samtools faidx'
                faidx_node = FastaIndexNode(reference)
                # Indexing of FASTA file using 'BuildSequenceDictionary.jar'
                dict_node = BuildSequenceDictNode(config=config,
                                                  reference=reference,
                                                  dependencies=(valid_node,))

                # Indexing of FASTA file using 'bwa index'
                bwa_node = BWAIndexNode(input_file=reference,
                                        dependencies=(valid_node,))
                # Indexing of FASTA file using ''
                bowtie2_node = Bowtie2IndexNode(input_file=reference,
                                                dependencies=(valid_node,))

                references[reference] = (valid_node, faidx_node, dict_node)
                references_bwa[reference] = (valid_node, faidx_node,
                                             dict_node, bwa_node)
                references_bowtie2[reference] = (valid_node, faidx_node,
                                                 dict_node, bowtie2_node)

            subdd["Nodes"] = references[reference]
            subdd["Nodes:BWA"] = references_bwa[reference]
            subdd["Nodes:Bowtie2"] = references_bowtie2[reference]


def run(config, args, pipeline_variant):
    if pipeline_variant not in ("bam", "trim"):
        raise ValueError("Unexpected BAM pipeline variant (%r)"
                         % (pipeline_variant,))

    if not os.path.exists(config.temp_root):
        try:
            os.makedirs(config.temp_root)
        except OSError, error:
            print_err("ERROR: Could not create temp root:\n\t%s" % (error,))
            return 1

    if not os.access(config.temp_root, os.R_OK | os.W_OK | os.X_OK):
        print_err("ERROR: Insufficient permissions for temp root: '%s'"
                  % (config.temp_root,))
        return 1

    # Init worker-threads before reading in any more data
    pipeline = Pypeline(config)

    try:
        print_info("Reading makefiles ...")
        makefiles = read_makefiles(config, args, pipeline_variant)
    except (MakefileError, paleomix.yaml.YAMLError, IOError), error:
        print_err("Error reading makefiles:",
                  "\n  %s:\n   " % (error.__class__.__name__,),
                  "\n    ".join(str(error).split("\n")))
        return 1

    logfile_template = time.strftime("bam_pipeline.%Y%m%d_%H%M%S_%%02i.log")
    paleomix.logger.initialize(config, logfile_template)
    logger = logging.getLogger(__name__)

    pipeline_func = build_pipeline_trimming
    if pipeline_variant == "bam":
        # Build .fai files for reference .fasta files
        index_references(config, makefiles)

        pipeline_func = build_pipeline_full

    print_info("Building BAM pipeline ", end='')
    for makefile in makefiles:
        # If a destination is not specified, save results in same folder as the
        # makefile
        filename = makefile["Statistics"]["Filename"]
        old_destination = config.destination
        if old_destination is None:
            config.destination = os.path.dirname(filename)

        try:
            nodes = pipeline_func(config, makefile)
        except paleomix.node.NodeError, error:
            logger.error("Error while building pipeline for '%s':\n%s",
                         filename, error)
            return 1

        config.destination = old_destination

        pipeline.add_nodes(*nodes)

    print_info("")

    if config.list_input_files:
        logger.info("Printing output files ...")
        pipeline.print_input_files()
        return 0
    elif config.list_output_files:
        logger.info("Printing output files ...")
        pipeline.print_output_files()
        return 0
    elif config.list_executables:
        logger.info("Printing required executables ...")
        pipeline.print_required_executables()
        return 0
    elif config.dot_file:
        logger.info("Writing dependency graph to %r ...", config.dot_file)
        if not pipeline.to_dot(config.dot_file):
            return 1
        return 0

    logger.info("Running BAM pipeline ...")
    if not pipeline.run(dry_run=config.dry_run,
                        max_threads=config.max_threads,
                        progress_ui=config.progress_ui):
        return 1

    return 0


def _print_usage(pipeline):
    basename = "%s_pipeline" % (pipeline,)
    usage = \
        "BAM Pipeline v{version}\n" \
        "Usage:\n" \
        "  -- {cmd} help           -- Display this message.\n" \
        "  -- {cmd} example [...]  -- Create example project.\n" \
        "  -- {cmd} makefile [...] -- Print makefile template.\n" \
        "  -- {cmd} dryrun [...]   -- Perform dry run of pipeline.\n" \
        "  -- {cmd} run [...]      -- Run pipeline on provided makefiles.\n" \
        "  -- {cmd} remap [...]    -- Re-map hits from previous alignment."

    print_info(usage.format(version=paleomix.__version__,
                            cmd=basename,
                            pad=" " * len(basename)))


def main(argv, pipeline="bam"):
    assert pipeline in ("bam", "trim"), pipeline

    commands = ("makefile", "mkfile", "run",
                "dry_run", "dry-run", "dryrun",
                "remap", "example", "examples")

    if not argv or (argv[0] == "help"):
        _print_usage(pipeline)
        return 0
    elif argv[0] not in commands:
        _print_usage(pipeline)
        return 1
    elif argv[0] in ("mkfile", "makefile"):
        return bam_mkfile.main(argv[1:], pipeline=pipeline)
    elif argv[0] in ("remap", "remap_prefix"):
        # Import here to avoid circular dependency issues
        import paleomix.tools.bam_pipeline.remap as bam_remap

        return bam_remap.main(argv[1:])
    elif argv[0] in ("example", "examples"):
        return paleomix.resources.copy_example("bam_pipeline", argv[1:])

    try:
        config, args = bam_config.parse_config(argv, pipeline)

        if not args[1:]:
            print_err("Please specify at least one makefile!")
            print_err("Use --help for more information.")
            return 1
        elif args and args[0].startswith("dry"):
            config.dry_run = True
    except bam_config.ConfigError, error:
        print_err(error)
        return 1

    return run(config, args[1:], pipeline_variant=pipeline)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import paleomix.common.versions as versions

from paleomix.atomiccmd.command import \
    AtomicCmd
from paleomix.node import \
    CommandNode, \
    Node, \
    NodeError

from paleomix.common.bedtools import \
    read_bed_file
from paleomix.common.fileutils import \
    move_file, \
    reroot_path


BEDTOOLS_VERSION \
    = versions.Requirement(call=("bedtools", "--version"),
                           search=r"bedtools v?(\d+)\.(\d+)\.(\d+)",
                           checks=versions.GE(2, 15, 0))


class SlopBedNode(CommandNode):
    def __init__(self, infile, outfile, genome, from_start=0, from_end=0,
                 strand_relative=False, dependencies=()):
        if type(from_start) != type(from_end):
            raise ValueError("Parameters 'from_start' and 'from_end' should "
                             "be of same type!")

        call = ["bedtools", "slop",
                "-i", "%(IN_FILE)s",
                "-g", "%(IN_GENOME)s",
                "-l", str(from_start),
                "-r", str(from_end)]

        if strand_relative:
            call.append("-s")
        if type(from_start) is float:
            call.append("-pct")

        command = AtomicCmd(call,
                            IN_FILE=infile,
                            IN_GENOME=genome,
                            OUT_STDOUT=outfile,
                            CHECK_VERSION=BEDTOOLS_VERSION)

        description = "<SlopBed: '%s' -> '%s'>" % (infile, outfile)

        CommandNode.__init__(self,
                             description=description,
                             command=command,
                             dependencies=dependencies)


class PaddedBedNode(Node):
    """Simple node for padding BED records a fixed amount and merging
    overlapping records. Columns beyond the 3rd column are dropped.
    """

    def __init__(self, infile, outfile, fai_file, amount=0, dependencies=()):
        self._amount = int(amount)
        self._infile = infile
        self._outfile = outfile
        self._fai_file = fai_file

        Node.__init__(self,
                      description='<PaddedBed (%i): %r -> %r>'
                      % (amount, infile, outfile),
                      input_files=(infile, fai_file),
                      output_files=(outfile,),
                      dependencies=dependencies)

    def _run(self, config, temp):
        contigs = {}
        with open(self._fai_file) as handle:
            for line in handle:
                name, length, _ = line.split('\t', 2)
                if name in contigs:
                    raise NodeError('Reference genome contains multiple '
                                    'identically named contigs (%r)!'
                                    % (name,))

                contigs[name] = int(length)

        with open(reroot_path(temp, self._outfile), 'w') as handle:
            for record in read_bed_file(self._infile, contigs=contigs):
                max_length = contigs[record.contig]
                record.start = max(0, record.start - self._amount)
                record.end = min(record.end + self._amount, max_length)

                handle.write('%s\n' % (record,))

    def _teardown(self, config, temp):
        source = reroot_path(temp, self._outfile)
        move_file(source, self._outfile)
